#include "esp_camera.h"
#include "Arduino.h"
#include <SPIFFS.h>
#include <WiFi.h>
#include <WebServer.h>
#include <ESP32_MailClient.h>
#include "time.h"

// Pin definition for CAMERA_MODEL_AI_THINKER
#define PWDN_GPIO_NUM       32
#define RESET_GPIO_NUM      -1
#define XCLK_GPIO_NUM        0
#define SIOD_GPIO_NUM       26
#define SIOC_GPIO_NUM       27

#define Y9_GPIO_NUM         35
#define Y8_GPIO_NUM         34
#define Y7_GPIO_NUM         39
#define Y6_GPIO_NUM         36
#define Y5_GPIO_NUM         21
#define Y4_GPIO_NUM         19
#define Y3_GPIO_NUM         18
#define Y2_GPIO_NUM          5
#define VSYNC_GPIO_NUM      25
#define HREF_GPIO_NUM       23
#define PCLK_GPIO_NUM       22

#define Ready_LED           33

WebServer server(80);
WebServer st_server(81);

const char *SSID = "your SSID";
const char *PASSWORD = "your password";

int cam_state[5]={0};
#define ope_stat            1

IPAddress ip(192, 168, 3, 200);         // IP Address
IPAddress gateway(192,168, 3, 1);       // Gateway Address
IPAddress subnet(255, 255, 255, 0);     // Subnet Address
IPAddress DNS(192, 168, 3, 1);          // Subnet Address

#define SMTP_HOST "smtp.gmail.com"
#define SMTP_PORT 465

/* The sign in credentials */
#define AUTHOR_EMAIL "xxxxxx@gmail.com"
#define AUTHOR_PASSWORD "yyyyyyyyyy"

/* Recipient's email*/
#define RECIPIENT_EMAIL "zzzzzzzzzz"

/* The SMTP Session object used for Email sending */
SMTPData smtpData;

const char* ntpServer = "pool.ntp.org";
const long  gmtOffset_sec = 9 * 3600;
const int   daylightOffset_sec = 0;

#define int_pin             GPIO_NUM_13

void IRAM_ATTR ISR_HANDLER(void* arg) 
{
    gpio_intr_disable(int_pin);
    digitalWrite (Ready_LED, !digitalRead(Ready_LED)) ;
    gpio_intr_enable(int_pin);
}

void setup() 
{
    sensor_t * cam_s;
    
    Serial.begin(115200);
    delay(100);

    Serial.println("Connecting to WiFi");
    WiFi.disconnect(true);
    WiFi.softAPdisconnect(true);
    delay(500);

    WiFi.mode(WIFI_STA);
    WiFi.config(ip, gateway, subnet, DNS);
    WiFi.begin(SSID, PASSWORD);
    delay(1000);

    // Try forever
    while (WiFi.status() != WL_CONNECTED) 
    {
        Serial.println("...Connecting to WiFi");
        delay(1000);
    }
    Serial.println("Connected");
    Serial.println(SSID);
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());

    server.begin();
    server.on("/", handleRoot);
    server.onNotFound(handleWebRequests);
    Serial.println("HTTP server started");

    st_server.begin();
    st_server.on("/", st_handleRoot);

    init_cam();
    cam_s = esp_camera_sensor_get();
    cam_state[ope_stat] = 0;
    cam_state[0] = cam_s->status.framesize;

    SPIFFS.begin();  

    //init and get the time
    configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);

    pinMode(Ready_LED, OUTPUT);
    digitalWrite (Ready_LED, LOW) ;

    pinMode(int_pin, INPUT_PULLDOWN);
    gpio_set_intr_type(int_pin, GPIO_INTR_POSEDGE);
    gpio_isr_handler_add(int_pin, ISR_HANDLER, NULL);
    gpio_intr_enable(int_pin);    
}

void loop() 
{
    server.handleClient();
    st_server.handleClient();
}

void send_mail()
{
    struct tm timeinfo;
    char m_title[80];
    
    //send email
    Serial.println("Sending email...");
    smtpData.setLogin("smtp.gmail.com", 465, AUTHOR_EMAIL, AUTHOR_PASSWORD);        //Set the Email host, port, account and password
    smtpData.setSender("ESP32-CAM", AUTHOR_EMAIL);                                  //Set the sender name and Email
    smtpData.setPriority("Normal");                                                 //Set Email priority or importance High, Normal, Low or 1 to 5 (1 is highest)
    getLocalTime(&timeinfo);                                                        
    sprintf(m_title,"%04d_%02d%02d_%02d:%02d:%02d",timeinfo.tm_year+1900,timeinfo.tm_mon+1,timeinfo.tm_mday,timeinfo.tm_hour,timeinfo.tm_min,timeinfo.tm_sec);
    smtpData.setSubject(m_title);                                                   //Set the subject
    smtpData.setMessage("Image captured and attached.", false);                     //Set the message - normal text or html format
    smtpData.addRecipient(RECIPIENT_EMAIL);                                         //Add recipients, can add more than one recipient                                
    smtpData.addAttachFile("/data.jpg");                                            //Add attach files from SD card
//    smtpData.setFileStorageType(MailClientStorageType::SD);
    smtpData.setFileStorageType(MailClientStorageType::SPIFFS);                     //Set the storage types to read the attach files (SD is default)
    smtpData.setSendCallback(sendCallback);
    
    if (!MailClient.sendMail(smtpData))                                             //Start sending Email, can be set callback function to track the status
        Serial.println("Error sending Email, " + MailClient.smtpErrorReason());
    smtpData.empty();                                                               //Clear all data from Email object to free memory
}

/* Callback function to get the Email sending status */
void sendCallback(SendStatus msg)
{
    //Print the current status
    Serial.println(msg.info());

    //Do something when complete
    if (msg.success())
    {
        Serial.println("----------------");
    }
}

void handleRoot() 
{
    String cmd;
    int a,fl;
    File dataFile;
    camera_fb_t * fb = NULL;
    sensor_t * cam_s;

    fl=1;
    cmd=server.argName(0);
    switch(cmd.toInt())
    {
      case 1: // Get Still
                fb = esp_camera_fb_get();
                dataFile = SPIFFS.open("/data.jpg", FILE_WRITE);
                dataFile.write(fb->buf, fb->len); // payload (image), payload length
                dataFile.close();
                esp_camera_fb_return(fb);
                Serial.println("Take a photo.");
                cam_state[ope_stat] = 2;
                break;

      case 2: //Set Camera Parameter
                Serial.println("Set Camera Parameter");
                cmd=server.arg("2");
                cam_state[0] = cmd.toInt();
                cam_s = esp_camera_sensor_get();
                cam_s->set_framesize(cam_s,(framesize_t)cam_state[0]);        //framesize
                break;

      case 3: //  Stream Start 
                cam_state[ope_stat] = 1;
                break;
                
      case 4: //  Stream Stop 
                Serial.println("Stream Stop.");
                cam_state[ope_stat] = 0;
                break;

      case 5: //  Send Mail
                send_mail();
                break;
                
      case 80: // Send camera parameter
                cmd="";
                for(a = 0; a < 2; a ++) cmd += (String(cam_state[a]) + ',');
                server.send(200, "text/plain", cmd);
                fl=0;
                break;
    }

    if(fl)
    {
        dataFile = SPIFFS.open("/monitor.html", FILE_READ);
        server.streamFile(dataFile,"text/html");
        dataFile.close();
    }
}

void init_cam()
{
    camera_config_t config;

    config.ledc_channel = LEDC_CHANNEL_0;
    config.ledc_timer = LEDC_TIMER_0;
    config.pin_d0 = Y2_GPIO_NUM;
    config.pin_d1 = Y3_GPIO_NUM;
    config.pin_d2 = Y4_GPIO_NUM;
    config.pin_d3 = Y5_GPIO_NUM;
    config.pin_d4 = Y6_GPIO_NUM;
    config.pin_d5 = Y7_GPIO_NUM;
    config.pin_d6 = Y8_GPIO_NUM;
    config.pin_d7 = Y9_GPIO_NUM;
    config.pin_xclk = XCLK_GPIO_NUM;
    config.pin_pclk = PCLK_GPIO_NUM;
    config.pin_vsync = VSYNC_GPIO_NUM;
    config.pin_href = HREF_GPIO_NUM;
    config.pin_sscb_sda = SIOD_GPIO_NUM;
    config.pin_sscb_scl = SIOC_GPIO_NUM;
    config.pin_pwdn = PWDN_GPIO_NUM;
    config.pin_reset = RESET_GPIO_NUM;
    config.xclk_freq_hz = 20000000;
    config.pixel_format = PIXFORMAT_JPEG;
    config.frame_size = FRAMESIZE_VGA; // FRAMESIZE_ + QVGA|CIF|VGA|SVGA|XGA|SXGA|UXGA
    config.jpeg_quality = 10;
    config.fb_count = 2;

    // Init Camera
    esp_camera_init(&config);
}

void st_handleRoot() 
{
    WiFiClient client;
    camera_fb_t * fb = NULL;

    Serial.println("HTTP st_handleRoot started");
    client = st_server.client();
    String response = "HTTP/1.1 200 OK\r\n";
    response += "Content-Type: multipart/x-mixed-replace; boundary=--frame\r\n\r\n";
    st_server.sendContent(response);
    while (1)
    {
      fb = esp_camera_fb_get();
      
      if (!client.connected())
      {
        esp_camera_fb_return(fb);
        break;
      }

      response = "--frame\r\n";
      response += "Content-Type: image/jpeg\r\n\r\n";
      st_server.sendContent(response);

      client.write(fb->buf, fb->len);
      st_server.sendContent("\r\n");
      esp_camera_fb_return(fb);

      if (!client.connected()) break;
    }
}

void handleWebRequests()
{
    String dataType = "text/plain";
    String path;
    File dataFile;
    camera_fb_t * fb;

    path = server.uri();
    if(path.endsWith(".txt")) dataType = "text/plain";
    else if(path.endsWith(".jpg")) dataType = "image/jpeg";
    else if(path.endsWith(".css")) dataType = "text/css";
    else if(path.endsWith(".js")) dataType = "application/javascript";
    else if(path.endsWith(".html")) dataType = "text/html";
    else if(path.endsWith(".ico")) dataType = "image/html";

    server.sendHeader("Cache-Control", "public, max-age=86400");
    
    dataFile = SPIFFS.open(path.c_str(), "r");
    server.streamFile(dataFile, dataType);
    dataFile.close();
    delay(5);
}
